@@ -65,7 +65,7 @@
             [puppetlabs.puppetdb.utils :as utils]
             [slingshot.slingshot :refer [try+ throw+]]
             [puppetlabs.puppetdb.command.constants
-             :refer [command-names supported-command-versions]]
+             :refer [command-names command-keys supported-command-versions]]
             [puppetlabs.trapperkeeper.services
              :refer [defservice service-context]]
             [schema.core :as s]
@@ -84,7 +84,7 @@
             [clj-time.coerce :as tcoerce]
             [puppetlabs.puppetdb.amq-migration :as mig]
             [puppetlabs.puppetdb.command.dlo :as dlo]
-            [overtone.at-at :refer [mk-pool stop-and-reset-pool! after]]
+            [overtone.at-at :as at-at :refer [mk-pool stop-and-reset-pool!]]
             [puppetlabs.puppetdb.threadpool :as gtp]
             [puppetlabs.puppetdb.utils.metrics :as mutils]))
 
@@ -270,10 +270,17 @@
   (let [id (str id)
         received-time (str (tcoerce/to-long received-time))
         duration (str (in-millis (interval start-time (now))))
-        command-name (command-names command-kw)]
-    (if-let [{:keys [puppet-version]} opts]
+        command-name (command-names command-kw)
+        puppet-version (:puppet-version opts)
+        obsolete-cmd? (:obsolete-cmd? opts)]
+    (cond
+      puppet-version
       (log/info (trs "[{0}-{1}] [{2} ms] ''{3}'' puppet v{4} command processed for {5}"
                      id received-time duration command-name puppet-version certname))
+      obsolete-cmd?
+      (log/info (trs "[{0}-{1}] [{2} ms] ''{3}'' command ignored (obsolete) for {4}"
+                     id received-time duration command-name certname))
+      :else
       (log/info (trs "[{0}-{1}] [{2} ms] ''{3}'' command processed for {4}"
                      id received-time duration command-name certname)))))
 
@@ -296,12 +303,23 @@
 
 ;; Fact replacement
 
+(defn rm-facts-by-regex [facts-blacklist fact-map]
+  (let [blacklisted? (fn [fact-name]
+                       (some #(re-matches % fact-name) facts-blacklist))]
+    (apply dissoc fact-map (filter blacklisted? (keys fact-map)))))
+
 (defn replace-facts*
-  [{:keys [payload id received] :as command} start-time db facts-blacklist]
+  [{:keys [payload id received] :as command} start-time db
+   {:keys [facts-blacklist facts-blacklist-type] :as blacklist-config}]
   (let [{:keys [certname package_inventory] :as fact-data} payload
         producer-timestamp (:producer_timestamp fact-data)
+        blacklisting? (seq blacklist-config)
+        rm-blacklisted (when blacklisting?
+                         (case facts-blacklist-type
+                           "regex" (partial rm-facts-by-regex facts-blacklist)
+                           "literal" #(apply dissoc % facts-blacklist)))
         trimmed-facts (cond-> fact-data
-                        (seq facts-blacklist) (update :values #(apply dissoc % facts-blacklist))
+                        blacklisting? (update :values rm-blacklisted)
                         (seq package_inventory) (update :package_inventory distinct))]
     (jdbc/with-transacted-connection' db :repeatable-read
       (scf-storage/maybe-activate-node! certname producer-timestamp)
@@ -309,12 +327,12 @@
     (log-command-processed-messsage id received start-time :replace-facts certname)))
 
 (defn replace-facts
-  [{:keys [payload version received] :as command} start-time db facts-blacklist]
+  [{:keys [payload version received] :as command} start-time db blacklist-config]
   (replace-facts* (upon-error-throw-fatality
                     (assoc command :payload (fact/normalize-facts version received payload)))
                   start-time
                   db
-                  facts-blacklist))
+                  blacklist-config))
 
 ;; Node deactivation
 
@@ -387,12 +405,12 @@
 (defn process-command!
   "Takes a command object and processes it to completion. Dispatch is
    based on the command's name and version information"
-  [{command-name :command version :version delete? :delete? :as command} db facts-blacklist]
+  [{command-name :command version :version delete? :delete? :as command} db blacklist-config]
   (when-not delete?
     (let [start (now)]
       (condp supported-command-version? [command-name version]
         "replace catalog" (replace-catalog command start db)
-        "replace facts" (replace-facts command start db facts-blacklist)
+        "replace facts" (replace-facts command start db blacklist-config)
         "store report" (store-report command start db)
         "deactivate node" (deactivate-node command start db)))))
 
@@ -420,7 +438,8 @@
 
 (defn make-cmd-processed-message [cmd ex]
   (conj
-   (conj (select-keys cmd [:id :command :version])
+   ;; :delete? from cmdref is needed to check when a command gets "bashed"
+   (conj (select-keys cmd [:id :command :version :delete?])
          [:producer-timestamp (get-in cmd [:payload :producer_timestamp])])
    (when ex
      [:exception ex])))
@@ -443,11 +462,11 @@
 (def quick-retry-count 4)
 
 (defn process-command-and-respond!
-  [{:keys [callback] :as cmd} db response-pub-chan stats-atom facts-blacklist]
+  [{:keys [callback] :as cmd} db response-pub-chan stats-atom blacklist-config]
   (try
     (let [result (call-with-quick-retry quick-retry-count
                   (fn []
-                    (process-command! cmd db facts-blacklist)))]
+                    (process-command! cmd db blacklist-config)))]
       (swap! stats-atom update :executed-commands inc)
       (callback {:command cmd :result result})
       (async/>!! response-pub-chan
@@ -481,8 +500,11 @@
   "Processes a command ref marked for deletion. This is similar to
   processing a non-delete cmdref except different metrics need to be
   updated to indicate the difference in command"
-  [{:keys [command version] :as cmdref} q scf-write-db response-chan stats]
-  (process-command-and-respond! cmdref scf-write-db response-chan stats)
+  [{:keys [command version certname id received] :as cmdref}
+   q scf-write-db response-chan stats blacklist-config]
+  (process-command-and-respond! cmdref scf-write-db response-chan stats blacklist-config)
+  (log-command-processed-messsage id received (now) (command-keys command)
+                                  certname {:obsolete-cmd? true})
   (queue/ack-command q {:entry (queue/cmdref->entry cmdref)})
   (update-counter! :depth command version dec!)
   (update-counter! :invalidated command version dec!))
@@ -491,7 +513,7 @@
   "Parses, processes and acknowledges a successful command ref and
   updates the relevant metrics. Any exceptions that arise are
   unhandled and expected to be caught by the caller."
-  [cmdref q scf-write-db response-chan stats facts-blacklist]
+  [cmdref q scf-write-db response-chan stats blacklist-config]
   (let [{:keys [command version] :as cmd} (queue/cmdref->cmd q cmdref)
         retries (count (:attempts cmdref))]
     (if-not cmd
@@ -512,7 +534,7 @@
          [(global-metric :processing-time)
           (cmd-metric command version :processing-time)]
 
-         (process-command-and-respond! cmd scf-write-db response-chan stats facts-blacklist)
+         (process-command-and-respond! cmd scf-write-db response-chan stats blacklist-config)
          (mark-both-metrics! command version :processed))
 
         (queue/ack-command q cmd)
@@ -520,21 +542,37 @@
 
 (def command-delay-ms (* 1000 60 60))
 
-(defn send-delayed-message
+;; For testing via with-redefs
+(defn enqueue-delayed-message [command-chan narrowed-entry]
+  (async/>!! command-chan narrowed-entry))
+
+;; For testing via with-redefs
+(def schedule-msg-after at-at/after)
+
+(defn schedule-delayed-message
   "Will delay `cmd` in the `delay-pool` threadpool for
   `command-delay-ms`. It will then be enqueued in `command-chan`
   for another attempt at processing"
-  [cmd ex command-chan delay-pool]
+  [cmd ex command-chan delay-pool stop-status]
   (let [narrowed-entry (-> cmd
                            (queue/cons-attempt ex)
                            (dissoc :payload))
-        {:keys [command version]} cmd]
+        {:keys [command version]} cmd
+        inc-msgs-if-not-stopping #(if (:stopping %)
+                                    %
+                                    (update % :executing-delayed inc))]
     (update-counter! :awaiting-retry command version inc!)
-    (after command-delay-ms
-           #(do
-              (async/>!! command-chan narrowed-entry)
-              (update-counter! :awaiting-retry command version dec!))
-           delay-pool)))
+    (schedule-msg-after
+     command-delay-ms
+     (fn []
+       (let [status (swap! stop-status inc-msgs-if-not-stopping)]
+         (when-not (:stopping status)
+           (try
+             (enqueue-delayed-message command-chan narrowed-entry)
+             (update-counter! :awaiting-retry command version dec!)
+             (finally
+               (swap! stop-status #(update % :executing-delayed dec)))))))
+     delay-pool)))
 
 (def ^:private iso-formatter (fmt-time/formatters :date-time))
 
@@ -542,7 +580,8 @@
   "Processes the message via (process-message msg), retrying messages
   that fail via (delay-message msg), and discarding messages that have
   fatal errors or have exceeded their maximum allowed attempts."
-  [q command-chan dlo delay-pool scf-write-db response-chan stats facts-blacklist]
+  [q command-chan dlo delay-pool scf-write-db response-chan stats
+   blacklist-config stop-status]
   (fn [{:keys [certname command version received delete? id] :as cmdref}]
     (try+
 
@@ -555,10 +594,10 @@
          (update! (cmd-metric command version :queue-time) q-time)))
 
      (if delete?
-       (process-delete-cmdref cmdref q scf-write-db response-chan stats)
+       (process-delete-cmdref cmdref q scf-write-db response-chan stats blacklist-config)
        (let [retries (count (:attempts cmdref))]
          (try+
-          (process-cmdref cmdref q scf-write-db response-chan stats facts-blacklist)
+          (process-cmdref cmdref q scf-write-db response-chan stats blacklist-config)
           (catch fatal? obj
             (mark! (global-metric :fatal))
             (let [ex (:cause obj)]
@@ -575,7 +614,7 @@
                    ex
                    (trs "[{0}] [{1}] Retrying after attempt {2} for {3}, due to: {4}"
                         id command retries certname ex))
-                  (send-delayed-message cmdref ex command-chan delay-pool))
+                  (schedule-delayed-message cmdref ex command-chan delay-pool stop-status))
                 (do
                   (log/error
                    ex
@@ -588,6 +627,8 @@
                   (trs "Fatal error parsing command: {0}" (:id cmdref)))
        (discard-message cmdref (:throwable &throw-context) q dlo)))))
 
+(def stop-commands-wait-ms (constantly 5000))
+
 (defn create-command-handler-threadpool
   "Creates an unbounded threadpool with the intent that access to the
   threadpool is bounded by the semaphore. Implicitly the threadpool is
@@ -600,7 +641,8 @@
 (defservice command-service
   PuppetDBCommandDispatcher
   [[:DefaultedConfig get-config]
-   [:PuppetDBServer shared-globals]]
+   [:PuppetDBServer shared-globals]
+   [:ShutdownService request-shutdown]]
   (init [this context]
     (let [response-chan (async/chan 1000)
           response-mult (async/mult response-chan)
@@ -615,54 +657,76 @@
              :response-chan response-chan
              :response-mult response-mult
              :response-chan-for-pub response-chan-for-pub
-             :response-pub (async/pub response-chan-for-pub :id))))
+             :response-pub (async/pub response-chan-for-pub :id)
+             ;; This coordination is needed until we no longer
+             ;; support jdk < 10.
+             ;; https://bugs.openjdk.java.net/browse/JDK-8176254
+             :stop-status (atom {:executing-delayed 0}))))
 
-  (start [this context]
-    (let [{:keys [scf-write-db dlo q]} (shared-globals)
-          {:keys [response-chan response-pub]} context
+  (start
+   [this context]
+   (let [config (get-config)
+         globals (shared-globals)
+         dlo (:dlo globals)
+         upgrade-q #(upgrade-activemq config (partial enqueue-command this) dlo)]
+     (if (get-in config [:global :upgrade-and-exit?])
+       (do
+         (upgrade-q)
+         ;; By this point we know that the pdb service has finished its
+         ;; start method because we depend on it for shared-globals,
+         ;; which means we've done everything that needs to be done.
+         (request-shutdown)
+         context)
+       (let [{:keys [command-chan scf-write-db q]} globals
+             {:keys [response-chan response-pub]} context
+             ;; From mq_listener
+             command-threadpool (create-command-handler-threadpool (conf/mq-thread-count config))
+             delay-pool (:delay-pool context)
+             handle-cmd (message-handler q
+                                         command-chan
+                                         dlo
+                                         delay-pool
+                                         scf-write-db
+                                         response-chan
+                                         (:stats context)
+                                         (-> (get-config)
+                                             :database
+                                             (select-keys [:facts-blacklist
+                                                           :facts-blacklist-type]))
+                                         (:stop-status context))]
 
-          ;; From mq_listener
-          command-threadpool (create-command-handler-threadpool (conf/mq-thread-count (get-config)))
-          {:keys [command-chan]} (shared-globals)
-          delay-pool (:delay-pool context)
-          handle-cmd (message-handler q
-                                      command-chan
-                                      dlo
-                                      delay-pool
-                                      scf-write-db
-                                      response-chan
-                                      (:stats context)
-                                      (get-in (get-config) [:database :facts-blacklist]))]
+         (doto (Thread. (fn []
+                          (try+ (gtp/dochan command-threadpool handle-cmd command-chan)
+                                ;; This only occurs when new work is submitted after the threadpool has
+                                ;; started shutting down, which means PDB itself is shutting down. The
+                                ;; command will be retried when PDB next starts up, so there's no reason to
+                                ;; tell the user about this by letting it bubble up until it's printed.
+                                (catch [:kind :puppetlabs.puppetdb.threadpool/rejected] _))))
+           (.setDaemon false)
+           (.start))
 
-      (doto (Thread. (fn []
-                       (try+ (gtp/dochan command-threadpool handle-cmd command-chan)
-                             ;; This only occurs when new work is submitted after the threadpool has
-                             ;; started shutting down, which means PDB itself is shutting down. The
-                             ;; command will be retried when PDB next starts up, so there's no reason to
-                             ;; tell the user about this by letting it bubble up until it's printed.
-                             (catch [:kind :puppetlabs.puppetdb.threadpool/rejected] _))))
-        (.setDaemon false)
-        (.start))
+         (upgrade-q)
+         (assoc context
+                :command-chan command-chan
+                :consumer-threadpool command-threadpool)))))
 
-      (upgrade-activemq (get-config)
-                        (partial enqueue-command this)
-                        dlo)
-
-      (assoc context
-             :command-chan command-chan
-             :consumer-threadpool command-threadpool)))
-
-  (stop [this {:keys [consumer-threadpool command-chan delay-pool] :as context}]
-    (async/close! command-chan)
-    (gtp/shutdown consumer-threadpool)
-    (when delay-pool
-      (stop-and-reset-pool! delay-pool))
+  (stop [this {:keys [stop-status consumer-threadpool command-chan delay-pool]
+               :as context}]
+    (some-> command-chan async/close!)
+    (some-> consumer-threadpool gtp/shutdown)
+    (some-> delay-pool stop-and-reset-pool!)
+    ;; Wait up to ~5s for https://bugs.openjdk.java.net/browse/JDK-8176254
+    (swap! stop-status #(assoc % :stopping true))
+    (if (utils/wait-for-ref-state stop-status (stop-commands-wait-ms)
+                                  #(= % {:stopping true :executing-delayed 0}))
+      (log/info (trs "Halted delayed command processsing"))
+      (log/info (trs "Forcibly terminating delayed command processing")))
     (async/unsub-all (:response-pub context))
     (async/untap-all (:response-mult context))
     (async/close! (:response-chan-for-pub context))
     (async/close! (:response-chan context))
-    (dissoc context :response-pub :response-chan :response-chan-for-pub :response-mult)
-    context)
+    (dissoc context
+            :response-pub :response-chan :response-chan-for-pub :response-mult))
 
   (stats [this]
     @(:stats (service-context this)))
