@@ -5,7 +5,8 @@
    The schemas in this file define what is expected to be present in the INI file
    and the format expected by the rest of the application."
   (:import [java.security KeyStore]
-           [org.joda.time Minutes Days Period])
+           [org.joda.time Minutes Days Period]
+           [java.util.regex PatternSyntaxException Pattern])
   (:require [puppetlabs.i18n.core :refer [trs]]
             [clojure.tools.logging :as log]
             [puppetlabs.kitchensink.core :as kitchensink]
@@ -72,8 +73,8 @@
      :log-statements (pls/defaulted-maybe String "true")
      :statements-cache-size (pls/defaulted-maybe s/Int 0)
      :connection-timeout (pls/defaulted-maybe s/Int 3000)
-     :facts-blacklist (s/conditional string? String
-                                     sequential? [s/Str])}))
+     :facts-blacklist pls/Blacklist
+     :facts-blacklist-type (pls/defaulted-maybe (s/enum "literal" "regex") "literal")}))
 
 (def write-database-config-in
   "Includes the common database config params, also the write-db specific ones"
@@ -107,7 +108,8 @@
    (s/optional-key :user) String
    (s/optional-key :password) String
    (s/optional-key :syntax_pgs) String
-   (s/optional-key :facts-blacklist) clojure.lang.PersistentVector})
+   (s/optional-key :facts-blacklist) clojure.lang.PersistentVector
+   :facts-blacklist-type String})
 
 (def write-database-config-out
   "Schema for parsed/processed database config that includes write database params"
@@ -203,13 +205,38 @@
   [{db-config :database :or {db-config {}} :as config}]
   (when (str/blank? (:subname db-config))
     (throw+
-     {:kind ::cli-error
-      :msg
+     {:type ::cli-error
+      :message
       (str "PuppetDB requires PostgreSQL."
            "  The [database] section must contain an appropriate"
            " \"//host:port/database\" subname setting.")}))
   config)
 
+(defn check-fact-regex [fact]
+  (try
+    (re-pattern fact)
+    (catch PatternSyntaxException e
+      (.getMessage e))))
+
+(defn convert-blacklist-config
+  "Validate and convert facts blacklist section of db-config to runtime format.
+  Throws a {:type ::cli-error :message m} exception describing errors when compiling
+  facts-blacklist regex patterns if :facts-blacklist-type is set to \"regex\"."
+  [{db-config :database :or {db-config {}} :as config}]
+  (if (= "regex" (:facts-blacklist-type db-config))
+    (let [patts (->> db-config
+                     :facts-blacklist
+                     pls/blacklist->vector
+                     (mapv check-fact-regex))]
+      (when-let [errors (seq (filter string? patts))]
+        (throw+
+         {:type ::cli-error
+          :message
+          (apply str "Unable to parse facts-blacklist patterns:\n"
+                 (interpose "\n" errors))}))
+      (assoc-in config [:database :facts-blacklist] patts))
+    config))
+
 (defn convert-section-config
   "validates and converts a `section-config` to `section-schema-out` using defaults from `section-schema-in`."
   [section-schema-in section-schema-out section-config]
@@ -415,12 +442,13 @@
 (def adjust-and-validate-tk-config
   (comp add-web-routing-service-config
         warn-retirements
-        validate-db-settings))
+        validate-db-settings
+        convert-blacklist-config))
 
 (defn hook-tk-parse-config-data
   "This is a robert.hooke compatible hook that is designed to intercept
    trapperkeeper configuration before it is used, so that we may munge &
-   customize it.  It may throw {:kind ::cli-error :msg m}."
+   customize it.  It may throw {:type ::cli-error :message m}."
   [f args]
   (adjust-and-validate-tk-config (f args)))
 
